/* Copyright (C) |Meso|Star> 2016 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "scpr.h"
#include "test_scpr_utils.h"

int
main(int argc, char** argv)
{
  const double coords[] = {
    0.0, 0.0,
    0.0, 0.5,
    0.0, 1.0,
    0.5, 0.0,
    0.5, 0.5,
    0.5, 1.0,
    1.0, 0.0,
    1.0, 0.5,
    1.0, 1.0
  };
  const size_t nverts = sizeof(coords)/sizeof(double[2]);
  const size_t indices[] = {
    0, 1, 3,
    3, 1, 4,
    1, 2, 4,
    4, 2, 5,
    3, 4, 6,
    6, 4, 7,
    4, 5, 7,
    7, 5, 8
  };
  const size_t ntris = sizeof(indices)/sizeof(size_t[3]);
  const size_t indices_bad[] = { 7, 5, 9 };
  size_t ids[3];
  double pos[2];
  size_t i, n;
  struct mem_allocator allocator;
  struct mesh_context ctx;
  struct scpr_mesh* mesh;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(scpr_mesh_create(NULL, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_create(&allocator, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_create(NULL, &mesh), RES_OK);

  CHECK(scpr_mesh_ref_get(NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_ref_get(mesh), RES_OK);
  CHECK(scpr_mesh_ref_put(NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_ref_put(mesh), RES_OK);
  CHECK(scpr_mesh_ref_put(mesh), RES_OK);

  CHECK(scpr_mesh_create(&allocator, &mesh), RES_OK);

  ctx.coords = coords;
  ctx.nverts = nverts;
  ctx.indices = indices;
  ctx.ntris = ntris;

  #define SETUP scpr_mesh_setup_indexed_vertices
  CHECK(SETUP(NULL, 0, NULL, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, NULL, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, NULL, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, NULL, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, get_ids, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, get_ids, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, get_ids, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, get_ids, 0, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, NULL, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, NULL, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, NULL, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, NULL, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, get_ids, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, get_ids, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, get_ids, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, get_ids, nverts, NULL, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, NULL, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, NULL, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, NULL, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, NULL, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, get_ids, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, get_ids, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, get_ids, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, get_ids, 0, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, NULL, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, NULL, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, NULL, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, NULL, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, 0, get_ids, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, 0, get_ids, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(NULL, ntris, get_ids, nverts, get_pos, &ctx), RES_BAD_ARG);
  CHECK(SETUP(mesh, ntris, get_ids, nverts, get_pos, &ctx), RES_OK);
  ctx.indices = indices_bad;
  CHECK(SETUP(mesh, 1, get_ids, nverts, get_pos, &ctx), RES_BAD_ARG);
  #undef SETUP

  CHECK(scpr_mesh_get_triangles_count(NULL, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_triangles_count(mesh, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_triangles_count(NULL, &n), RES_BAD_ARG);
  CHECK(scpr_mesh_get_triangles_count(mesh, &n), RES_OK);
  CHECK(n, 0);

  CHECK(scpr_mesh_get_vertices_count(NULL, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_vertices_count(mesh, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_vertices_count(NULL, &n), RES_BAD_ARG);
  CHECK(scpr_mesh_get_vertices_count(mesh, &n), RES_OK);
  CHECK(n, 0);

  ctx.indices = indices;
  CHECK(scpr_mesh_setup_indexed_vertices
    (mesh, ntris, get_ids, nverts, get_pos, &ctx), RES_OK);
  CHECK(scpr_mesh_get_triangles_count(mesh, &n), RES_OK);
  CHECK(n, ntris);
  CHECK(scpr_mesh_get_vertices_count(mesh, &n), RES_OK);
  CHECK(n, nverts);

  CHECK(scpr_mesh_get_indices(NULL, ntris, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(mesh, ntris, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(NULL, 0, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(mesh, 0, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(NULL, ntris, ids), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(mesh, ntris, ids), RES_BAD_ARG);
  CHECK(scpr_mesh_get_indices(NULL, 0, ids), RES_BAD_ARG);
  FOR_EACH(i, 0, ntris) {
    CHECK(scpr_mesh_get_indices(mesh, i, ids), RES_OK);
    CHECK(ids[0], indices[i*3+0]);
    CHECK(ids[1], indices[i*3+1]);
    CHECK(ids[2], indices[i*3+2]);
  }

  CHECK(scpr_mesh_get_position(NULL, nverts, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(mesh, nverts, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(NULL, 0, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(mesh, 0, NULL), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(NULL, nverts, pos), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(mesh, nverts, pos), RES_BAD_ARG);
  CHECK(scpr_mesh_get_position(NULL, 0, pos), RES_BAD_ARG);
  FOR_EACH(i, 0, nverts) {
    CHECK(scpr_mesh_get_position(mesh, i, pos), RES_OK);
    CHECK(pos[0], coords[i*2+0]);
    CHECK(pos[1], coords[i*2+1]);
  }

  CHECK(scpr_mesh_ref_put(mesh), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}

