/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "ssf.h"
#include "test_ssf_utils.h"

#include <rsys/double3.h>

int
main(int argc, char** argv)
{
  const size_t NSTEPS = 100000;
  struct mem_allocator allocator;
  struct ssp_rng* rng;
  struct ssf_bsdf* bsdf;
  struct ssf_bxdf* bxdf;
  struct ssf_bxdf* bxdf2;
  struct ssf_fresnel* fresnel;
  double wo[3], wi[3], pdf;
  double N[3];
  double E, V, SE, R;
  double sum, sqr_sum;
  double sample[4];
  size_t i;
  int type;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);
  CHECK(ssp_rng_create(&allocator, &ssp_rng_threefry, &rng), RES_OK);

  CHECK(ssf_bsdf_create(NULL, NULL), RES_BAD_ARG);
  CHECK(ssf_bsdf_create(&allocator, NULL), RES_BAD_ARG);
  CHECK(ssf_bsdf_create(NULL, &bsdf), RES_OK);
  CHECK(ssf_fresnel_create(&allocator, &ssf_fresnel_no_op, &fresnel), RES_OK);

  CHECK(ssf_bsdf_ref_get(NULL), RES_BAD_ARG);
  CHECK(ssf_bsdf_ref_get(bsdf), RES_OK);
  CHECK(ssf_bsdf_ref_put(NULL), RES_BAD_ARG);
  CHECK(ssf_bsdf_ref_put(bsdf), RES_OK);
  CHECK(ssf_bsdf_ref_put(bsdf), RES_OK);

  CHECK(ssf_bsdf_create(&allocator, &bsdf), RES_OK);
  CHECK(ssf_bxdf_create(&allocator, &ssf_specular_reflection, &bxdf), RES_OK);
  CHECK(ssf_bxdf_create(&allocator, &ssf_specular_reflection, &bxdf2), RES_OK);
  CHECK(ssf_specular_reflection_setup(bxdf, fresnel), RES_OK);
  CHECK(ssf_specular_reflection_setup(bxdf2, fresnel), RES_OK);

  CHECK(ssf_bsdf_add(NULL, NULL, -1.0), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(bsdf, NULL, -1.0), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(NULL, bxdf, -1.0), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(bsdf, bxdf, -1.0), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(NULL, NULL, 1), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(bsdf, NULL, 1), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(NULL, bxdf, 1), RES_BAD_ARG);
  CHECK(ssf_bsdf_add(bsdf, bxdf, 1), RES_OK);

  d3_normalize(wo, d3(wo, 1, 1, 0));
  d3(N, 0.0, 1.0, 0.0);
  R = ssf_bsdf_sample(bsdf, rng, wo, N, wi, &type, &pdf);
  CHECK(type, SSF_SPECULAR | SSF_REFLECTION);
  CHECK(eq_eps(R, ssf_fresnel_eval(fresnel, d3_dot(wi, N)), 1.e-6), 1);
  CHECK(ssf_bsdf_eval(bsdf, wo, N, wi), 0);
  CHECK(ssf_bsdf_pdf(bsdf, wo, N, wi), 0);

  CHECK(ssf_specular_reflection_setup(bxdf, fresnel), RES_OK);
  R = ssf_bsdf_sample(bsdf, rng, wo, N, wi, &type, &pdf);
  CHECK(eq_eps(R, ssf_fresnel_eval(fresnel, d3_dot(wi, N)), 1.e-6), 1);
  CHECK(type, SSF_SPECULAR | SSF_REFLECTION);

  CHECK(ssf_bsdf_add(bsdf, bxdf2, 1), RES_OK);

  /* Estimate the BSDF rho. FIXME the following computation is certainly
   * totally wrong */
  sum = 0;
  sqr_sum = 0;
  ssp_ran_hemisphere_uniform(rng, N, sample);
  d3_set(wo, sample);
  FOR_EACH(i, 0, NSTEPS) {
    double weight = ssf_bsdf_sample(bsdf, rng, wo, N, wi, &type, &pdf);
    CHECK(type, SSF_SPECULAR | SSF_REFLECTION);
    CHECK(ssf_bsdf_eval(bsdf, wo, N, wi), 0);
    CHECK(ssf_bsdf_pdf(bsdf, wo, N, wi), 0);
    sum += weight;
    sqr_sum += weight*weight;
  }
  E =  sum / (double)NSTEPS;
  V = sqr_sum / (double)NSTEPS - E*E;
  SE = sqrt(V / (double)NSTEPS);
  CHECK(eq_eps(1, E, SE), 1);

  CHECK(ssf_bsdf_clear(NULL), RES_BAD_ARG);
  CHECK(ssf_bsdf_clear(bsdf), RES_OK);

  CHECK(ssf_bsdf_sample(bsdf, rng, wo, N, wi, &type, &pdf), 0.0);
  CHECK(type, 0);
  CHECK(pdf, 0.0);

  CHECK(ssf_bsdf_ref_put(bsdf), RES_OK);
  CHECK(ssf_bxdf_ref_put(bxdf), RES_OK);
  CHECK(ssf_bxdf_ref_put(bxdf2), RES_OK);
  CHECK(ssf_fresnel_ref_put(fresnel), RES_OK);
  CHECK(ssp_rng_ref_put(rng), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}

